from genson import SchemaBuilder
from json import dump, load
from jsonschema import validate
import requests


def get_json_from_url(url):
    return requests.get(
        url=url, params={}, headers={"Cache-Control": "no-cache"}
    ).json()


def get_json_from_file(filename):
    with open(filename) as f:
        return load(f)


def get_json(uri):
    return (
        get_json_from_url(uri)
        if uri.lower().startswith("http://") or uri.lower().startswith("https://")
        else get_json_from_file(uri)
    )


def save_json(data, filename):
    with open(filename, "w") as f:
        dump(data, f, indent=4, sort_keys=True)


def download_json(url, filename):
    save_json(get_json_from_url(url), filename)


def jsons_to_schema(jsons, dollar_schema="https://library.cdisc.org/api/mdr/schema"):
    builder = SchemaBuilder(dollar_schema)
    for version in jsons:
        builder.add_object(version)
    schema = builder.to_schema()
    for version in jsons:
        validate(instance=version, schema=schema)
    return schema


def merge_jsons_to_schema(schema, jsons):
    """
    Take a schema and list of json documents and return a new schema with additional items

    Keyword arguments:
    schema -- json schema
    jsons -- list of new json data
    """
    builder = SchemaBuilder()
    builder.add_schema(schema)
    for version in jsons:
        builder.add_object(version)
    return builder.to_schema()


from collections import Counter
from enum import Enum, auto
from json import dumps


class DetailLevel(Enum):
    BASIC = auto()
    ADVANCED = auto()


def create_counter(key_dictlist: dict) -> Counter:
    """Converts {key, dictlist} to Counter(key, len(dictlist))"""
    return Counter({key: len(dict_list) for key, dict_list in key_dictlist.items()})


def simple_text(value: str) -> str:
    return [{"text": value}]


def json_text(value, indent: int = None) -> str:
    return simple_text(dumps(obj=value, sort_keys=True, indent=indent))


def enumerate_any(map_or_sequence):
    return (
        map_or_sequence.items()
        if ObjectTypes.is_mapping_type(map_or_sequence)
        else enumerate(map_or_sequence)
    )


class ObjectTypes(Enum):
    SEQUENCE_TYPES = {"collapsed": "[...]", "types": (list, tuple, range)}
    MAPPING_TYPES = {"collapsed": "{...}", "types": (dict,)}
    SET_TYPES = {"collapsed": "[...]", "types": (set, frozenset)}
    RECURSIVE_TYPES = {
        "collapsed": "...",
        "types": SEQUENCE_TYPES["types"] + MAPPING_TYPES["types"] + SET_TYPES["types"],
    }

    def contains(self, obj) -> bool:
        return type(obj) in self.value["types"]

    def collapsed(self) -> str:
        return self.value["collapsed"]

    @staticmethod
    def is_mapping_type(obj) -> bool:
        return ObjectTypes.MAPPING_TYPES.contains(obj)

    @staticmethod
    def is_sequence_type(obj) -> bool:
        return ObjectTypes.SEQUENCE_TYPES.contains(obj)

    @staticmethod
    def is_recursive_type(obj) -> bool:
        return ObjectTypes.RECURSIVE_TYPES.contains(obj)



import unittest
from tests.helpers import compare_with_file


class TestLogicBranches(unittest.TestCase):
    def test_composite_key(self):
        compare_with_file("composite_key")

    def test_dict_drop_str(self):
        compare_with_file("dict_drop_str")

    def test_dict_add_str(self):
        compare_with_file("dict_add_str")

    def test_dict_drop_list(self):
        compare_with_file("dict_drop_list")

    def test_dict_add_list(self):
        compare_with_file("dict_add_list")

    def test_duplicates(self):
        compare_with_file("duplicates")

    def test_list_drop_dict(self):
        compare_with_file("list_drop_dict")

    def test_list_add_dict(self):
        compare_with_file("list_add_dict")

    def test_list_drop_item(self):
        compare_with_file("list_drop_item")

    def test_list_add_item(self):
        compare_with_file("list_add_item")

    def test_list_item_type(self):
        compare_with_file("list_item_type")

    def test_minor_major(self):
        compare_with_file("minor_major")

    def test_multikey(self):
        compare_with_file("multikey")

    def test_text(self):
        compare_with_file("text")

    def test_value(self):
        compare_with_file("value")


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from ..models.diff_text import DiffText
from re import sub
from ..utils.helpers import (
    DetailLevel,
    ObjectTypes,
    enumerate_any,
    json_text,
    simple_text,
)


class DiffItemBuilder:
    def __init__(self, src_dict: dict, dynamic_headers: list):
        self.src_dict = src_dict
        self.dynamic_headers = dynamic_headers
        self.diff_item = {}

    def build_action(self):
        """ What type of diff action was performed? """
        self.diff_item["Action"] = simple_text(self.src_dict["action"])

    def build_func(self):
        """ Which logical branch was reached? """
        self.diff_item["func"] = self.src_dict["func"]

    def build_path(self):
        """ Full path of current JSON node """
        self.diff_item["path"] = self.src_dict["path"].path_list()

    def build_attribute_path(self):
        """ Generate Filtered JSON-Path string """
        self.diff_item["Attribute Path"] = simple_text(
            self.src_dict["path"].get_attribute_jsonpath()
        )

    def build_location_path(self):
        """ Generate Unfiltered JSON-Path string """
        self.diff_item["Location Path"] = simple_text(
            self.src_dict["path"].get_location_jsonpath()
        )

    def build_change_level(self):
        pass

    def build_opcodes(self):
        pass

    def build_value_previous(self):
        pass

    def build_value_updated(self):
        pass

    def build_attribute_previous(self):
        pass

    def build_attribute_updated(self):
        pass

    def build_impact(self):
        pass

    def build_dynamic_columns(self):
        pass


class AddDropBuilder(DiffItemBuilder):
    """ Add or Drop - Populate previous if old item dropped or populate updated if new item added """


class UpdateBuilder(DiffItemBuilder):
    """ Value update - Determine string indices for highlighting differences """

    @staticmethod
    def is_minor_change(old: str, new: str) -> bool:
        """
        Take 2 strings. Returns True if the only difference between the two strings is case or non-alphanumeric changes.

        Keyword arguments:
        old -- Old string to compare
        new -- New string to compare
        """
        return sub("[^0-9A-Z]+", "", old.upper()) == sub("[^0-9A-Z]+", "", new.upper())

    def __init__(self, src_dict: dict, dynamic_headers: list):
        super().__init__(src_dict, dynamic_headers)
        self.diff_text = DiffText(self.src_dict["old"], self.src_dict["new"])

    def build_change_level(self):
        self.diff_item["Change Level"] = simple_text(
            "Minor"
            if UpdateBuilder.is_minor_change(self.src_dict["old"], self.src_dict["new"])
            else "Major"
        )

    def build_opcodes(self):
        self.diff_item["opcodes"] = self.diff_text.opcodes

    def build_value_previous(self):
        self.diff_item["Value (previous)"] = json_text(self.diff_text.old_val)

    def build_value_updated(self):
        self.diff_item["Value (updated)"] = json_text(self.diff_text.new_val)

    def build_attribute_previous(self):
        self.diff_item["Attribute (previous)"] = self.diff_text.old_str

    def build_attribute_updated(self):
        self.diff_item["Attribute (updated)"] = self.diff_text.new_str


class BasicBuilder(DiffItemBuilder):
    def build_impact(self):
        """ Impact is the leaf node in the Path """
        self.diff_item["Impact"] = simple_text(
            self.src_dict["path"].get_alias()
            if self.src_dict["path"].get_alias()
            else self.src_dict["path"].get_name()
        )

    def build_dynamic_columns(self):
        """ Generate dynamic columns """
        for dynamic_column in self.src_dict["path"].get_dynamic_column_aliases():
            alias = dynamic_column.get("alias", dynamic_column["name"])
            if alias not in self.dynamic_headers:
                self.dynamic_headers[alias] = {"names": [dynamic_column["name"]]}
            elif dynamic_column["name"] not in self.dynamic_headers[alias]["names"]:
                self.dynamic_headers[alias]["names"].append(dynamic_column["name"])
            if "alias" in dynamic_column:
                self.dynamic_headers[alias]["alias"] = dynamic_column["alias"]
            if "description" in dynamic_column:
                self.dynamic_headers[alias]["description"] = dynamic_column[
                    "description"
                ]
            self.diff_item[alias] = simple_text(dynamic_column["value"])


class AdvancedBuilder(DiffItemBuilder):
    def build_impact(self):
        """ Impact is the leaf node in the Path """
        self.diff_item["Impact"] = simple_text(self.src_dict["path"].get_name())

    def build_dynamic_columns(self):
        """ Generate dynamic columns """
        for dynamic_column in self.src_dict["path"].get_dynamic_column_names():
            alias = dynamic_column.get("alias", dynamic_column["name"])
            if alias not in self.dynamic_headers:
                self.dynamic_headers[alias] = {"names": [dynamic_column["name"]]}
            elif dynamic_column["name"] not in self.dynamic_headers[alias]["names"]:
                self.dynamic_headers[alias]["names"].append(dynamic_column["name"])
            if "alias" in dynamic_column:
                self.dynamic_headers[alias]["alias"] = dynamic_column["alias"]
            if "description" in dynamic_column:
                self.dynamic_headers[alias]["description"] = dynamic_column[
                    "description"
                ]
            self.diff_item[dynamic_column["name"]] = simple_text(
                dynamic_column["value"]
            )


class BasicAddDropBuilder(AddDropBuilder, BasicBuilder):
    def build_recursive_type(self, attr_name: str, attr_value, target_key: str):
        attr_copy = deepcopy(attr_value)
        has_simple_items = False
        for key, value in enumerate_any(attr_copy):
            if ObjectTypes.is_mapping_type(value):
                # dict inside list/dict
                attr_copy[key] = ObjectTypes.MAPPING_TYPES.collapsed()
            elif ObjectTypes.is_sequence_type(value):
                # list inside list/dict
                attr_copy[key] = ObjectTypes.SEQUENCE_TYPES.collapsed()
            else:
                # simple type inside list/dict
                has_simple_items = True
        self.diff_item[target_key] = json_text(
            attr_copy if has_simple_items or not attr_name else attr_name, 4
        )

    def build_simple_type(self, attr_value, target_key: str):
        self.diff_item[target_key] = simple_text(attr_value)

    def build_attribute(self, source_name_key: str, source_key: str, target_key: str):
        attr_name = self.src_dict.get(source_name_key)
        attr_value = self.src_dict.get(source_key)
        if attr_value is not None and ObjectTypes.is_recursive_type(attr_value):
            self.build_recursive_type(attr_name, attr_value, target_key)
        elif attr_value is not None:
            self.build_simple_type(attr_value, target_key)

    def build_attribute_previous(self):
        self.build_attribute(
            "old_name",
            "old",
            "Attribute (previous)",
        )

    def build_attribute_updated(self):
        self.build_attribute(
            "new_name",
            "new",
            "Attribute (updated)",
        )


class AdvancedAddDropBuilder(AddDropBuilder, AdvancedBuilder):
    def build_attribute(self, source_key: str, target_key: str):
        attribute = self.src_dict.get(source_key)
        if attribute is not None and ObjectTypes.is_recursive_type(attribute):
            self.diff_item[target_key] = json_text(attribute, 4)
        elif attribute is not None:
            self.diff_item[target_key] = simple_text(attribute)

    def build_attribute_previous(self):
        self.build_attribute("old", "Attribute (previous)")

    def build_attribute_updated(self):
        self.build_attribute("new", "Attribute (updated)")


class BasicUpdateBuilder(UpdateBuilder, BasicBuilder):
    pass


class AdvancedUpdateBuilder(UpdateBuilder, AdvancedBuilder):
    pass


class DiffItemDirector:
    @staticmethod
    def get_diff_item(
        src_dict: dict, dynamic_headers: list, detail_level: DetailLevel
    ) -> dict:
        diff_item_builder = None
        old = src_dict.get("old", None)
        new = src_dict.get("new", None)
        is_update = (
            old is not None
            and type(old) == str
            and new is not None
            and type(new) == str
        )
        if detail_level == DetailLevel.BASIC and is_update:
            diff_item_builder = BasicUpdateBuilder(src_dict, dynamic_headers)
        elif detail_level == DetailLevel.BASIC and not is_update:
            diff_item_builder = BasicAddDropBuilder(src_dict, dynamic_headers)
        elif detail_level == DetailLevel.ADVANCED and is_update:
            diff_item_builder = AdvancedUpdateBuilder(src_dict, dynamic_headers)
        elif detail_level == DetailLevel.ADVANCED and not is_update:
            diff_item_builder = AdvancedAddDropBuilder(src_dict, dynamic_headers)

        diff_item_builder.build_action()
        diff_item_builder.build_func()
        diff_item_builder.build_path()
        diff_item_builder.build_attribute_path()
        diff_item_builder.build_location_path()
        diff_item_builder.build_change_level()
        diff_item_builder.build_opcodes()
        diff_item_builder.build_value_previous()
        diff_item_builder.build_value_updated()
        diff_item_builder.build_attribute_previous()
        diff_item_builder.build_attribute_updated()
        diff_item_builder.build_impact()
        diff_item_builder.build_dynamic_columns()

        return diff_item_builder.diff_item



from difflib import SequenceMatcher


class DiffText:
    """Represents a diff between a pair of text"""

    FORMATS = {
        "REPLACE_FORMAT": {"bold": True, "color": "#FF3300"},  # Orange-Red
        "BORDER_FORMAT": {"bold": True, "color": "#00CC55"},  # Blue-Green
    }

    FORMAT = "format"
    MARKER = "marker"
    TEXT = "text"

    OPCODE_CASES = {
        "equal": "_opcode_equal",
        "replace": "_opcode_replace",
        "delete": "_opcode_delete",
        "insert": "_opcode_insert",
    }

    def __init__(
        self,
        old: any = None,
        new: any = None,
    ):
        # Value update - Determine string indices for highlighting differences
        opcodes = SequenceMatcher(
            None,
            old,
            new,
            False,
        ).get_opcodes()
        self.opcodes = opcodes
        self.old_val = []
        self.new_val = []
        # These 2 variables are lists of dicts and markers.
        # dict - Each dict contains a chunk of text and, if it is not plaintext, a format that applies to that chunk of text.
        # marker - This is a string that represents a location where text was removed or dropped in the diff pair.
        self.old_str = []
        self.new_str = []
        for opcode in opcodes:
            getattr(self, DiffText.OPCODE_CASES[opcode[0]])(
                old[opcode[1] : opcode[2]],
                new[opcode[3] : opcode[4]],
            )
        DiffText._highlight_surrounding(self.old_str)
        DiffText._highlight_surrounding(self.new_str)
        self.old_str = DiffText._combine_adjacent_same_format(self.old_str)
        self.new_str = DiffText._combine_adjacent_same_format(self.new_str)

    def _opcode_equal(
        self,
        old_substr: str,
        new_substr: str,
    ):
        self.old_str.append({DiffText.TEXT: old_substr})
        self.new_str.append({DiffText.TEXT: new_substr})

    def _opcode_replace(
        self,
        old_substr: str,
        new_substr: str,
    ):
        self.old_str.append(
            {
                DiffText.TEXT: old_substr,
                DiffText.FORMAT: DiffText.FORMATS["REPLACE_FORMAT"],
            }
        )
        self.new_str.append(
            {
                DiffText.TEXT: new_substr,
                DiffText.FORMAT: DiffText.FORMATS["REPLACE_FORMAT"],
            }
        )
        self.old_val.append(old_substr)
        self.new_val.append(new_substr)

    def _opcode_delete(
        self,
        old_substr: str,
        new_substr: str,
    ):
        self.old_str.append(
            {
                DiffText.TEXT: old_substr,
                DiffText.FORMAT: DiffText.FORMATS["REPLACE_FORMAT"],
            }
        )
        self.new_str.append(DiffText.MARKER)
        if new_substr:
            self.new_str.append(
                {
                    DiffText.TEXT: new_substr,
                }
            )
        self.old_val.append(old_substr)
        self.new_val.append(new_substr)

    def _opcode_insert(
        self,
        old_substr: str,
        new_substr: str,
    ):
        self.old_str.append(DiffText.MARKER)
        if old_substr:
            self.old_str.append(
                {
                    DiffText.TEXT: old_substr,
                }
            )
        self.new_str.append(
            {
                DiffText.TEXT: new_substr,
                DiffText.FORMAT: DiffText.FORMATS["REPLACE_FORMAT"],
            }
        )
        self.old_val.append(old_substr)
        self.new_val.append(new_substr)

    @staticmethod
    def _highlight_single_char(
        texts: list, relative_marker_index: int, string: str, non_space_index: int
    ):
        replacements = []
        if non_space_index > 0:
            replacements.append({DiffText.TEXT: string[:non_space_index]})
        replacements.append(
            {
                DiffText.FORMAT: DiffText.FORMATS["BORDER_FORMAT"],
                DiffText.TEXT: string[non_space_index : non_space_index + 1],
            }
        )
        if non_space_index < len(string) - 1:
            replacements.append({DiffText.TEXT: string[non_space_index + 1 :]})
        texts[relative_marker_index : relative_marker_index + 1] = replacements

    @staticmethod
    def _highlight_surrounding(texts: list):
        marker_index = texts.index(DiffText.MARKER) if DiffText.MARKER in texts else -1
        while marker_index >= 0:
            # Highlight "after" index first, so that we don't mess up the index when we go to highlight "before" index
            relative_marker_index = marker_index + 1
            while (
                relative_marker_index < len(texts)
                and texts[relative_marker_index] != DiffText.MARKER
            ):
                text = texts[relative_marker_index]
                string = text[DiffText.TEXT]
                if not string.isspace():
                    if DiffText.FORMAT not in text:
                        # Replace the proceeding text with a highlighted string
                        DiffText._highlight_single_char(
                            texts,
                            relative_marker_index,
                            string,
                            len(string) - len(string.lstrip()),
                        )
                    break
                relative_marker_index += 1
            # Highlight "before" index
            relative_marker_index = marker_index - 1
            while (
                relative_marker_index >= 0
                and texts[relative_marker_index] != DiffText.MARKER
            ):
                text = texts[relative_marker_index]
                string = text[DiffText.TEXT]
                if not string.isspace():
                    if DiffText.FORMAT not in text:
                        # Replace the preceeding text with a highlighted string
                        DiffText._highlight_single_char(
                            texts,
                            relative_marker_index,
                            string,
                            len(string.rstrip()) - 1,
                        )
                    break
                relative_marker_index -= 1
            # Prepare for next iteration
            texts.remove(DiffText.MARKER)
            marker_index = (
                texts.index(DiffText.MARKER) if DiffText.MARKER in texts else -1
            )

    @staticmethod
    def _combine_adjacent_same_format(texts: list) -> list:
        """
        Take a list of formatted text chunks. If two adjacent text chunks have the same format, concatenate the text chunks together into a single chunk with the same format.
        Returns the combined list. len(combined) <= len(texts)

        Keyword arguments:
        texts -- List of dicts. Each dict contains a chunk of text and, if it is not plaintext, a format that applies to that chunk of text.
        """
        combined = []
        for text_index, text in enumerate(texts):
            if text_index == 0 or text.get(DiffText.FORMAT, None) != texts[
                text_index - 1
            ].get(DiffText.FORMAT, None):
                combined.append(text)
            else:
                combined[-1][DiffText.TEXT] += text[DiffText.TEXT]
        return combined



import unittest
from keyediffer.models.diff_item import DiffItemDirector
from keyediffer.utils.helpers import DetailLevel
from keyediffer.models.diff_path import DiffPath
from tests.helpers import compare_objects


class TestDiffItems(unittest.TestCase):
    def test_basic_add(self):
        static_data = {
            "Action": [{"text": "Add"}],
            "func": "Test",
            "path": [{"alias": "Root Alias", "name": "root"}],
            "Attribute Path": [{"text": "$['root']"}],
            "Location Path": [{"text": "$['root']"}],
            "Attribute (updated)": [{"text": "New Value"}],
            "Impact": [{"text": "Root Alias"}],
        }
        dynamic_data = DiffItemDirector.get_diff_item(
            {
                "new": "New Value",
                "action": "Add",
                "func": "Test",
                "path": DiffPath(
                    parent=DiffPath(),
                    schema={"properties": {"root": {"alias": "Root Alias"}}},
                    name="root",
                ),
            },
            (),
            DetailLevel.BASIC,
        )
        compare_objects(
            static_data,
            dynamic_data,
        )

    def test_basic_drop(self):
        static_data = {
            "Action": [{"text": "Drop"}],
            "func": "Test",
            "path": [{"alias": "Root Alias", "name": "root"}],
            "Attribute Path": [{"text": "$['root']"}],
            "Location Path": [{"text": "$['root']"}],
            "Attribute (previous)": [{"text": "Old Value"}],
            "Impact": [{"text": "Root Alias"}],
        }
        dynamic_data = DiffItemDirector.get_diff_item(
            {
                "old": "Old Value",
                "action": "Drop",
                "func": "Test",
                "path": DiffPath(
                    parent=DiffPath(),
                    schema={"properties": {"root": {"alias": "Root Alias"}}},
                    name="root",
                ),
            },
            (),
            DetailLevel.BASIC,
        )
        compare_objects(
            static_data,
            dynamic_data,
        )

    def test_basic_update(self):
        static_data = {
            "Action": [{"text": "Value Update"}],
            "func": "Test",
            "path": [{"alias": "Root Alias", "name": "root"}],
            "Attribute Path": [{"text": "$['root']"}],
            "Location Path": [{"text": "$['root']"}],
            "Change Level": [{"text": "Major"}],
            "opcodes": [("replace", 0, 1, 0, 1)],
            "Value (previous)": [{"text": '["A"]'}],
            "Value (updated)": [{"text": '["B"]'}],
            "Attribute (previous)": [
                {"text": "A", "format": {"bold": True, "color": "#FF3300"}}
            ],
            "Attribute (updated)": [
                {"text": "B", "format": {"bold": True, "color": "#FF3300"}}
            ],
            "Impact": [{"text": "Root Alias"}],
        }
        dynamic_data = DiffItemDirector.get_diff_item(
            {
                "old": "A",
                "new": "B",
                "action": "Value Update",
                "func": "Test",
                "path": DiffPath(
                    parent=DiffPath(),
                    schema={"properties": {"root": {"alias": "Root Alias"}}},
                    name="root",
                ),
            },
            (),
            DetailLevel.BASIC,
        )
        compare_objects(
            static_data,
            dynamic_data,
        )

    def test_advanced_add(self):
        static_data = {
            "Action": [{"text": "Add"}],
            "func": "Test",
            "path": [{"alias": "Root Alias", "name": "root"}],
            "Attribute Path": [{"text": "$['root']"}],
            "Location Path": [{"text": "$['root']"}],
            "Attribute (updated)": [{"text": "New Value"}],
            "Impact": [{"text": "root"}],
        }
        dynamic_data = DiffItemDirector.get_diff_item(
            {
                "new": "New Value",
                "action": "Add",
                "func": "Test",
                "path": DiffPath(
                    parent=DiffPath(),
                    schema={"properties": {"root": {"alias": "Root Alias"}}},
                    name="root",
                ),
            },
            (),
            DetailLevel.ADVANCED,
        )
        compare_objects(
            static_data,
            dynamic_data,
        )

    def test_advanced_drop(self):
        static_data = {
            "Action": [{"text": "Drop"}],
            "func": "Test",
            "path": [{"alias": "Root Alias", "name": "root"}],
            "Attribute Path": [{"text": "$['root']"}],
            "Location Path": [{"text": "$['root']"}],
            "Attribute (previous)": [{"text": "Old Value"}],
            "Impact": [{"text": "root"}],
        }
        dynamic_data = DiffItemDirector.get_diff_item(
            {
                "old": "Old Value",
                "action": "Drop",
                "func": "Test",
                "path": DiffPath(
                    parent=DiffPath(),
                    schema={"properties": {"root": {"alias": "Root Alias"}}},
                    name="root",
                ),
            },
            (),
            DetailLevel.ADVANCED,
        )
        compare_objects(
            static_data,
            dynamic_data,
        )

    def test_advanced_update(self):
        static_data = {
            "Action": [{"text": "Value Update"}],
            "func": "Test",
            "path": [{"alias": "Root Alias", "name": "root"}],
            "Attribute Path": [{"text": "$['root']"}],
            "Location Path": [{"text": "$['root']"}],
            "Change Level": [{"text": "Major"}],
            "opcodes": [("replace", 0, 1, 0, 1)],
            "Value (previous)": [{"text": '["A"]'}],
            "Value (updated)": [{"text": '["B"]'}],
            "Attribute (previous)": [
                {"text": "A", "format": {"bold": True, "color": "#FF3300"}}
            ],
            "Attribute (updated)": [
                {"text": "B", "format": {"bold": True, "color": "#FF3300"}}
            ],
            "Impact": [{"text": "root"}],
        }
        dynamic_data = DiffItemDirector.get_diff_item(
            {
                "old": "A",
                "new": "B",
                "action": "Value Update",
                "func": "Test",
                "path": DiffPath(
                    parent=DiffPath(),
                    schema={"properties": {"root": {"alias": "Root Alias"}}},
                    name="root",
                ),
            },
            (),
            DetailLevel.ADVANCED,
        )
        compare_objects(
            static_data,
            dynamic_data,
        )


if __name__ == "__main__":
    unittest.main()



from collections import deque


class DiffPath:
    def __init__(self, parent=None, schema: dict = None, name: str = None):
        self.parent = parent
        self.properties = {"name": name} if name else {}
        if schema and schema.get("properties", {}).get(name, {}).get("alias", {}):
            self.properties["alias"] = schema["properties"][name]["alias"]
        if schema and schema.get("properties", {}).get(name, {}).get("description", {}):
            self.properties["description"] = schema["properties"][name]["description"]

    def set_filter(self, schema: dict, key: dict, doc: dict):
        self.properties["filters"] = DiffPath._get_path_filter(schema, key, doc)

    def path_list(self) -> list:
        return (self.parent.path_list() if self.parent else []) + (
            [self.properties] if self.properties else []
        )

    def get_location_jsonpath(self) -> str:
        location = "$"
        for subpath in self.path_list():
            location += DiffPath.jsonpath_field_name(subpath["name"])
        return location

    def get_attribute_jsonpath(self) -> str:
        location = "$"
        for subpath in self.path_list():
            location += DiffPath.jsonpath_field_name(subpath["name"])
            if "filters" in subpath:
                location += DiffPath.get_filter_jsonpath(subpath["filters"])
        return location

    def get_leaf_jsonpath(self) -> str:
        return f"${DiffPath.get_filter_jsonpath(self.properties['filters'])}"

    def get_name(self) -> str:
        return self.properties["name"]

    def get_alias(self) -> str:
        return self.properties.get("alias", None)

    def get_values_from_key(self) -> list:
        return [
            fltr["value"]
            for fltr in DiffPath._filters_bfs_sorted(self.properties["filters"])
            if "value" in fltr
        ]

    def get_dynamic_column_names(self) -> list:
        return self._get_dynamic_columns(
            lambda fltr, dynamic_path: dynamic_path
            + DiffPath._get_jsonpath_to_trunk(fltr)
        )

    def get_dynamic_column_aliases(self) -> list:
        return self._get_dynamic_columns(
            lambda fltr, dynamic_path: fltr["alias"]
            if "alias" in fltr
            else dynamic_path + DiffPath._get_jsonpath_to_trunk(fltr)
        )

    def __lt__(self, other) -> bool:
        return self.get_leaf_jsonpath() < other.get_leaf_jsonpath()

    def __eq__(self, other) -> bool:
        return self.get_leaf_jsonpath() == other.get_leaf_jsonpath()

    def __hash__(self) -> int:
        return hash(self.get_leaf_jsonpath())

    @staticmethod
    def escape_jsonpath_field(field: str) -> str:
        field = field.replace("\\", "\\\\").replace("'", "\\'")
        return f"'{field}'"

    @staticmethod
    def jsonpath_field_name(name: str) -> str:
        return f"[{DiffPath.escape_jsonpath_field(name)}]"

    @staticmethod
    def get_filter_jsonpath(fltr: dict) -> str:
        return f"[?(@{'&@'.join(DiffPath._get_filter_jsonpath(fltr))})]"

    @staticmethod
    def _filters_bfs_sorted(fltrs: dict):
        # Uses a queue to return key filters in BFS order.
        # Uses sorting to return keys in alpha order at each level
        filter_queue = deque([{"filters": fltrs}])
        while filter_queue:
            head = filter_queue.popleft()
            for _, filter_value in sorted(head["filters"].items()):
                if "parent" in head:
                    filter_value["parent"] = head["parent"]
                yield filter_value
                if "filters" in filter_value:
                    filter_queue.append(
                        {"parent": filter_value, "filters": filter_value["filters"]}
                    )

    @staticmethod
    def _get_path_filter(schema: dict, key: dict, doc: dict) -> dict:
        path = {}
        for name, value in key.items():
            path[name] = {"name": name}
            if schema.get("properties", {}).get(name, {}).get("alias", {}):
                path[name]["alias"] = schema["properties"][name]["alias"]
            if schema.get("properties", {}).get(name, {}).get("description", {}):
                path[name]["description"] = schema["properties"][name]["description"]
            if value:
                path[name]["filters"] = DiffPath._get_path_filter(
                    schema.get("properties", {}).get(name, {}),
                    value,
                    doc[name],
                )
            else:
                path[name]["value"] = doc[name]
        return path

    @staticmethod
    def _get_filtered_jsonpath_to_trunk(fltr: dict) -> str:
        return (
            (
                DiffPath._get_filtered_jsonpath_to_trunk(fltr["parent"])
                if "parent" in fltr
                else ""
            )
            + DiffPath.jsonpath_field_name(fltr["name"])
            + (
                f"={DiffPath.escape_jsonpath_field(fltr['value'])}"
                if "value" in fltr
                else ""
            )
        )

    @staticmethod
    def _get_filter_jsonpath(fltrs: dict) -> list:
        return [
            DiffPath._get_filtered_jsonpath_to_trunk(fltr)
            for fltr in DiffPath._filters_bfs_sorted(fltrs)
            if "filters" not in fltr and "value" in fltr
        ]

    @staticmethod
    def _get_jsonpath_to_trunk(fltr: dict) -> str:
        return (
            DiffPath._get_jsonpath_to_trunk(fltr["parent"]) if "parent" in fltr else ""
        ) + DiffPath.jsonpath_field_name(fltr["name"])

    def _get_dynamic_columns(self, get_column_path) -> list:
        dynamic_columns = []
        dynamic_path = "$"
        for subpath in self.path_list():
            dynamic_path += DiffPath.jsonpath_field_name(subpath["name"])
            if "filters" in subpath and subpath["filters"]:
                for fltr in DiffPath._filters_bfs_sorted(subpath["filters"]):
                    if "filters" not in fltr and "value" in fltr:
                        dynamic_columns.append(
                            {
                                "name": get_column_path(fltr, dynamic_path),
                                "value": fltr["value"],
                            }
                        )
                        if "alias" in fltr:
                            dynamic_columns[-1]["alias"] = fltr["alias"]
                        if "description" in fltr:
                            dynamic_columns[-1]["description"] = fltr["description"]
        return dynamic_columns



from io import BytesIO
from json import dumps
from xlsxwriter import Workbook


def format_xlsx_sheet(
    json_sheet,
    worksheet,
    cell_format,
    MIN_COL_WIDTH=5,
    MAX_COL_WIDTH=100,
    FILTER_BUTTON_WIDTH=3,
):
    col_widths = [MIN_COL_WIDTH] * len(json_sheet["head"])
    for head_index, head in enumerate(json_sheet["head"]):
        col_widths[head_index] = max(col_widths[head_index], len(head))
    for row in json_sheet["body"]:
        for head_index, head in enumerate(json_sheet["head"]):
            if head in row:
                length = 0
                for string_part in row[head]:
                    length += len(string_part["text"])
                col_widths[head_index] = max(col_widths[head_index], length)
    for col_index, col_width in enumerate(col_widths):
        worksheet.set_column(
            first_col=col_index,
            last_col=col_index,
            width=min(MAX_COL_WIDTH, col_width + FILTER_BUTTON_WIDTH),
            cell_format=cell_format,
        )
    worksheet.freeze_panes(1, 0)
    if json_sheet["head"]:
        worksheet.autofilter(0, 0, len(json_sheet["body"]), len(json_sheet["head"]) - 1)


def write_formatted_xlsx_cells(
    json_sheet,
    workbook,
    worksheet,
):
    formats = {}
    for row_index, row in enumerate(json_sheet["body"]):
        for col_index, col in enumerate(json_sheet["head"]):
            if col in row:
                string_parts = []
                for string_part in row[col]:
                    if "format" in string_part:
                        format_json = dumps(obj=string_part["format"], sort_keys=True)
                        fmt = (
                            formats[format_json]
                            if format_json in formats
                            else workbook.add_format(string_part["format"])
                        )
                        formats[format_json] = fmt
                        string_parts.append(fmt)
                    string_parts.append(string_part["text"])
                if len(string_parts) == 1:
                    worksheet.write(row_index + 1, col_index, string_parts[0])
                elif len(string_parts) == 2:
                    worksheet.write(
                        row_index + 1, col_index, string_parts[1], string_parts[0]
                    )
                elif len(string_parts) >= 2:
                    worksheet.write_rich_string(row_index + 1, col_index, *string_parts)


def update_xlsx_workbook(data: list, workbook: Workbook) -> Workbook:
    """Take a json list of data sheets and an Excel Workbook, build the workbook contents from the data, and return the same workbook."""
    text_wrap_format = workbook.add_format()
    text_wrap_format.set_text_wrap()
    for json_sheet in data:
        worksheet = workbook.add_worksheet(json_sheet["title"])
        format_xlsx_sheet(json_sheet, worksheet, text_wrap_format)
        worksheet.write_row(0, 0, json_sheet["head"])
        write_formatted_xlsx_cells(json_sheet, workbook, worksheet)
    workbook.close()
    return workbook


def save_xlsx(data: list, filename: str) -> Workbook:
    """Take a json list of data sheets and filename, create an Excel file, and return the created workbook."""
    return update_xlsx_workbook(data, Workbook(filename))


def create_xlsx_stream(data: list) -> BytesIO:
    """Take a json list of data sheets, create an Excel workbook in memory, and return the created stream."""
    output = BytesIO()
    update_xlsx_workbook(data, Workbook(output, {"in_memory": True}))
    output.seek(0)
    return output


from json import dumps
import unittest
from jsonpath_ng.ext import parse
from keyediffer.utils.json_utils import get_json
from tests.helpers import run_diff


class TestJSONPath(unittest.TestCase):
    @staticmethod
    def jsonpath_compare(jsonpath, row, field_name, filename):
        if field_name in row:
            static_data = "".join(part["text"] for part in row[field_name])
            dynamic_data = [
                match.value
                for match in jsonpath.find(get_json(f"tests/data/{filename}.json"))
            ]
            assert (
                len(dynamic_data) == 1 and static_data == dynamic_data[0]
            ), f"\n{dumps(static_data, sort_keys=True)}\n{dumps(dynamic_data, sort_keys=True)}"

    def test_jsonpath(self):
        prefix = "value"
        diffs = run_diff(prefix)
        advanced = next(filter(lambda item: item["title"] == "Advanced", diffs))["body"]
        for row in advanced:
            jsonpath = parse(row["Attribute Path"][0]["text"])
            TestJSONPath.jsonpath_compare(
                jsonpath, row, "Attribute (previous)", f"{prefix}_previous"
            )
            TestJSONPath.jsonpath_compare(
                jsonpath, row, "Attribute (updated)", f"{prefix}_updated"
            )


if __name__ == "__main__":
    unittest.main()



# Built-in imports
from collections import Counter, OrderedDict
from copy import copy, deepcopy
from datetime import datetime
import json
from importlib.resources import files

# Installed imports
from jsonschema import validate, ValidationError

# Source imports
from .models.diff_item import DiffItemDirector
from .models.diff_path import DiffPath
from .utils.helpers import create_counter, DetailLevel, simple_text, ObjectTypes
from .utils.json_utils import jsons_to_schema


def simplify_json_schema(schema):
    schema.pop("required", None)
    schema.pop("type", None)
    if "properties" in schema:
        for prop in schema["properties"]:
            simplify_json_schema(schema["properties"][prop])
    if "items" in schema:
        if "properties" in schema["items"] and "key" not in schema["items"]:
            schema["items"]["key"] = [{"": {}}]
        simplify_json_schema(schema["items"])
    return schema


def create_schema_template(jsons):
    return simplify_json_schema(jsons_to_schema(jsons))


def get_doc_id(schema, doc):
    if "doc_id" in schema:
        return doc[schema["doc_id"]]
    elif type(doc) == dict:
        return "".join(
            get_doc_id(schema["properties"][prop], doc[prop]) for prop in sorted(doc)
        )
    elif type(doc) in (list, tuple):
        return "".join(get_doc_id(schema["items"], item) for item in doc)


def get_validation_features(doc_id, issue, details):
    return {
        "Document": simple_text(doc_id),
        "Issue": simple_text(issue),
        "Details": simple_text(details),
    }


def validate_doc_keys_exist(schema_item: dict, doc_dict: dict) -> bool:
    """
    Take a schema item, and document dict and recursively check that all keys in the schema item have a corresponding value in the document dict.
    Returns True if dict has all key values, otherwise returns False

    Keyword arguments:
    schema_item -- Schema item that corresponds to the doc_dict
    doc_dict -- dict in the json document
    """
    valid = True
    for name, value in sorted(schema_item.items()):
        if name in doc_dict:
            if value:
                valid &= validate_doc_keys_exist(schema_item[name], doc_dict[name])
        else:
            valid = False
    return valid


def validate_key_aliases(schema: dict, key: dict, alias_set: set) -> list:
    duplicates = []
    for name, value in sorted(key.items()):
        if value:
            duplicates += validate_key_aliases(
                schema["properties"][name], value, alias_set
            )
        elif "alias" in schema["properties"][name]:
            alias = schema["properties"][name]["alias"]
            if alias in alias_set:
                duplicates += [alias]
            else:
                alias_set.add(alias)
    return duplicates


def custom_json_validate(doc_id: str, schema: dict, doc, alias_set: set) -> list:
    """
    Take a schema and json document, and recursively return a list of validation errors, if any.

    Keyword arguments:
    doc_id -- json document identfier
    schema -- json schema subtree
    doc -- json document subtree
    alias_set -- a set of alias names for dynamic columns along the path to the current node
    """
    exceptions = []
    alias_set = copy(alias_set)
    if type(doc) not in (dict, list, tuple, int, str, float, bool, type(None)):
        exceptions.append(
            get_validation_features(
                doc_id,
                "Invalid JSON type in document",
                f"type={str(type(doc))}, value={str(doc)}",
            )
        )
    if type(doc) == dict:
        for dict_key, dict_value in sorted(doc.items()):
            exceptions += custom_json_validate(
                doc_id,
                schema["properties"][dict_key]
                if schema
                and "properties" in schema
                and dict_key in schema["properties"]
                else None,
                dict_value,
                alias_set,
            )
    elif type(doc) in (list, tuple):
        types = {type(item) for item in doc}
        if len(types) == 1:
            if dict in types:
                if not schema or "items" not in schema:
                    exceptions.append(
                        get_validation_features(
                            doc_id,
                            "Missing schema subtree for list of dicts.",
                            json.dumps(obj=doc, sort_keys=True),
                        )
                    )
                elif "key" not in schema["items"]:
                    exceptions.append(
                        get_validation_features(
                            doc_id,
                            "Missing schema key attribute for list of dicts.",
                            json.dumps(obj=doc, sort_keys=True),
                        )
                    )
                elif not (schema["items"]["key"]):
                    exceptions.append(
                        get_validation_features(
                            doc_id,
                            "Missing schema key attribute value for list of dicts.",
                            json.dumps(obj=doc, sort_keys=True),
                        )
                    )
                else:
                    for key in schema["items"]["key"]:
                        for item in validate_key_aliases(
                            schema["items"], key, alias_set
                        ):
                            exceptions.append(
                                get_validation_features(
                                    doc_id,
                                    "Duplicate key alias.",
                                    f"alias={item}",
                                )
                            )
                    for item in doc:
                        for key in schema["items"]["key"]:
                            if not (validate_doc_keys_exist(key, item)):
                                exceptions.append(
                                    get_validation_features(
                                        doc_id,
                                        "Document list item missing a key value.",
                                        f"key={schema['items']['key']}, item={json.dumps(obj=item, sort_keys=True)}",
                                    )
                                )
            elif len(types & {list, tuple}) == 1:
                exceptions.append(
                    get_validation_features(
                        doc_id,
                        "Diffing does not support multi-dimensional lists.",
                        json.dumps(obj=doc, sort_keys=True),
                    )
                )
        elif len(types) > 1:
            exceptions.append(
                get_validation_features(
                    doc_id,
                    "Diffing does not support lists of items with mixed types.",
                    json.dumps(obj=doc, sort_keys=True),
                )
            )
        for item in doc:
            exceptions += custom_json_validate(
                doc_id,
                schema["items"] if schema and "items" in schema else None,
                item,
                alias_set,
            )
    return exceptions


def json_validate(schema, doc):
    exceptions = []
    doc_id = get_doc_id(schema, doc)
    if not (doc_id):
        exceptions.append(
            get_validation_features("", "doc_id is missing from schema or document.")
        )
    try:
        # Validate json against json schema
        validate(schema=schema, instance=doc)
    except ValidationError as err:
        exceptions.append(
            get_validation_features(doc_id, "Schema Validation Error.", err)
        )
    exceptions += custom_json_validate(doc_id, schema, doc, set())
    return exceptions


def parse_dict(schema, path, old, new, is_excluded):
    # Property in both dicts
    for key in sorted(old.keys() & new.keys()):
        yield from traverse(
            schema.get("properties", {}).get(key, {}),
            DiffPath(path, schema, key),
            old[key],
            new[key],
            is_excluded or key in schema.get("exclusions", {}),
        )
    # Property has been dropped in new dict
    for key in sorted(old.keys() - new.keys()):
        is_sequence_type = ObjectTypes.is_sequence_type(old[key])
        yield {
            "action": "Drop",
            "func": "dict",
            "path": DiffPath(path, schema, key),
            "old_name": key,
            "old": old[key],
            "is_excluded": is_excluded
            or key in schema.get("exclusions", {})
            or is_sequence_type,
        }
        # If a sequence has been dropped, list all values within the sequence
        if is_sequence_type:
            yield from traverse(
                schema.get("properties", {}).get(key, {}),
                DiffPath(path, schema, key),
                old[key],
                [],
                is_excluded or key in schema.get("exclusions", {}),
            )
    # Property has been added in new dict
    for key in sorted(new.keys() - old.keys()):
        is_sequence_type = ObjectTypes.is_sequence_type(new[key])
        yield {
            "action": "Add",
            "func": "dict",
            "path": DiffPath(path, schema, key),
            "new_name": key,
            "new": new[key],
            "is_excluded": is_excluded
            or key in schema.get("exclusions", {})
            or is_sequence_type,
        }
        # If a sequence has been added, list all values within the sequence
        if is_sequence_type:
            yield from traverse(
                schema.get("properties", {}).get(key, {}),
                DiffPath(path, schema, key),
                [],
                new[key],
                is_excluded or key in schema.get("exclusions", {}),
            )


def create_path_dictlist_map(
    schema: dict, path: DiffPath, dict_list: list, key: dict
) -> dict:
    path_dictlist_map = {}
    for item in dict_list:
        filtered_path = deepcopy(path)
        filtered_path.set_filter(schema["items"], key, item)
        path_dictlist_map.setdefault(filtered_path, []).append(item)
    return path_dictlist_map


def compare_old_new_dicts(schema, path, old, new, is_excluded):
    old_difference = copy(old)
    new_difference = copy(new)
    for key_tree in schema["items"]["key"]:
        old_path_dictlist_map = create_path_dictlist_map(
            schema, path, old_difference, key_tree
        )
        new_path_dictlist_map = create_path_dictlist_map(
            schema, path, new_difference, key_tree
        )
        old_counter = create_counter(old_path_dictlist_map)
        new_counter = create_counter(new_path_dictlist_map)
        old_difference = [
            item
            for key in (old_path_dictlist_map.keys() - new_path_dictlist_map.keys())
            for item in old_path_dictlist_map[key]
        ]
        new_difference = [
            item
            for key in (new_path_dictlist_map.keys() - old_path_dictlist_map.keys())
            for item in new_path_dictlist_map[key]
        ]
        # Dict in both lists
        for key in sorted(old_path_dictlist_map.keys() & new_path_dictlist_map.keys()):
            # n^2 complexity, but if we use good keys, n (# of duplicates) should be 1
            for old_item in old_path_dictlist_map[key]:
                for new_item in new_path_dictlist_map[key]:
                    yield from parse_dict(
                        schema["items"],
                        key,
                        old_item,
                        new_item,
                        is_excluded,
                    )
            # Duplicate Dict has been dropped in new list
            if old_counter[key] > new_counter[key]:
                yield {
                    "action": "Drop",
                    "func": "list-dict",
                    "path": key,
                    "old_name": ", ".join(key.get_values_from_key()),
                    "old": old_path_dictlist_map[key][0]
                    if len(old_path_dictlist_map[key]) == 1
                    else old_path_dictlist_map[key],
                    "is_excluded": is_excluded,
                }
            # Duplicate Dict has been added in new list
            if new_counter[key] > old_counter[key]:
                yield {
                    "action": "Add",
                    "func": "list-dict",
                    "path": key,
                    "new_name": ", ".join(key.get_values_from_key()),
                    "new": new_path_dictlist_map[key][0]
                    if len(new_path_dictlist_map[key]) == 1
                    else new_path_dictlist_map[key],
                    "is_excluded": is_excluded,
                }
    old_path_dictlist_map = create_path_dictlist_map(
        schema, path, old_difference, schema["items"]["key"][0]
    )
    new_path_dictlist_map = create_path_dictlist_map(
        schema, path, new_difference, schema["items"]["key"][0]
    )
    old_counter = create_counter(old_path_dictlist_map)
    new_counter = create_counter(new_path_dictlist_map)
    # Dict has been dropped in new list
    for key in sorted((old_counter - new_counter).keys()):
        yield {
            "action": "Drop",
            "func": "list-dict",
            "path": key,
            "old_name": ", ".join(key.get_values_from_key()),
            "old": old_path_dictlist_map[key][0]
            if len(old_path_dictlist_map[key]) == 1
            else old_path_dictlist_map[key],
            "is_excluded": is_excluded,
        }
    # Dict has been added in new list
    for key in sorted((new_counter - old_counter).keys()):
        yield {
            "action": "Add",
            "func": "list-dict",
            "path": key,
            "new_name": ", ".join(key.get_values_from_key()),
            "new": new_path_dictlist_map[key][0]
            if len(new_path_dictlist_map[key]) == 1
            else new_path_dictlist_map[key],
            "is_excluded": is_excluded,
        }


def compare_old_new_values(schema, path, old, new, is_excluded):
    old_counter = Counter(old)
    new_counter = Counter(new)
    # Value has been dropped in new list
    for val in sorted((old_counter - new_counter).elements()):
        yield {
            "action": "Drop",
            "func": "list-nodict",
            "path": path,
            "old": val,
            "is_excluded": is_excluded,
        }
    # Value has been added in new list
    for val in sorted((new_counter - old_counter).elements()):
        yield {
            "action": "Add",
            "func": "list-nodict",
            "path": path,
            "new": val,
            "is_excluded": is_excluded,
        }


def parse_list(schema, path, old, new, is_excluded):
    old_item_type = type(old[0]) if old else None
    new_item_type = type(new[0]) if new else None
    if not old_item_type or not new_item_type or old_item_type == new_item_type:
        if dict in (old_item_type, new_item_type):
            yield from compare_old_new_dicts(schema, path, old, new, is_excluded)
        else:
            yield from compare_old_new_values(schema, path, old, new, is_excluded)
    else:
        # Items in list have a different type in new list
        yield {
            "action": "Type Update",
            "func": "list-unequal",
            "path": path,
            "old": old,
            "new": new,
            "is_excluded": is_excluded,
        }


def parse_value(schema, path, old, new, is_excluded):
    if old != new:
        yield {
            "action": "Value Update",
            "func": "value",
            "path": path,
            "old": old,
            "new": new,
            "is_excluded": is_excluded,
        }


def traverse(schema, path, old, new, is_excluded):
    """Traverse a pair of json files and a corresponding schema and return a list of differences.

    Keyword arguments:
    schema -- json schema
    path -- list of path parts and filters that point to the current path in the json files
    old -- previous json file for comparison
    new -- updated json file for comparison
    is_excluded -- whether or not the diff item should be excluded from the basic view
    """
    if type(old) == type(new):
        if type(old) == dict:
            yield from parse_dict(schema, path, old, new, is_excluded)
        elif type(old) in (list, tuple):
            yield from parse_list(schema, path, old, new, is_excluded)
        else:
            yield from parse_value(schema, path, old, new, is_excluded)
    else:
        yield {
            "action": "Type Update",
            "func": "traverse",
            "path": path,
            "old": old,
            "new": new,
            "is_excluded": is_excluded,
        }


PRE_BASIC_HEADERS = [
    "Updated Version",
    "Previous Version",
    "Action",
    "Impact",
    "Change Level",
]
POST_BASIC_HEADERS = ["Attribute (updated)", "Attribute (previous)"]
PRE_ADVANCED_HEADERS = ["Attribute Path"]
POST_ADVANCED_HEADERS = [
    "Location Path",
    "Action",
    "Impact",
    "Change Level",
    "Attribute (updated)",
    "Attribute (previous)",
    "Value (updated)",
    "Value (previous)",
]
INFO_HEADERS = ["Attribute", "Value"]
VALIDATION_HEADERS = ["Document", "Issue", "Details"]


def set_readme_body(
    readme_tables: list,
    readme_name: str,
    insert_before: str,
    dynamic_body: list,
    diffs: list,
):
    readme_body = next(
        filter(lambda table: table["title"] == readme_name, readme_tables)
    )["body"]
    insert_position = readme_body.index(
        next(
            filter(
                lambda column: column["Column Name"][0]["text"] == insert_before,
                readme_body,
            )
        )
    )
    readme_body[insert_position:insert_position] = dynamic_body
    # Populate "Impact" list
    next(
        filter(lambda column: column["Column Name"][0]["text"] == "Impact", readme_body)
    )["Description"][0]["text"] += ", ".join(
        sorted(set([f'"{diff_item["Impact"][0]["text"]}"' for diff_item in diffs]))
    )


def load_readme_tables(
    dynamic_basic_headers: OrderedDict,
    dynamic_advanced_headers: OrderedDict,
    basic_diffs: list,
    advanced_diffs: list,
) -> list:
    readme_tables = json.loads(
        files(__name__).joinpath("resources", "readme_tables.json").read_bytes()
    )
    # Basic Readme
    dynamic_body = []
    for dynamic_header_alias, dynamic_header_value in dynamic_basic_headers.items():
        dynamic_body.append({"Column Name": simple_text(dynamic_header_alias)})
        if "description" in dynamic_header_value:
            dynamic_body[-1]["Description"] = simple_text(
                dynamic_header_value["description"]
            )
    set_readme_body(
        readme_tables,
        "ReadMe for Basic Tab",
        (POST_BASIC_HEADERS + PRE_BASIC_HEADERS)[0],
        dynamic_body,
        basic_diffs,
    )
    # Advanced Readme
    dynamic_body = []
    for dynamic_header_alias, dynamic_header_value in dynamic_advanced_headers.items():
        for dynamic_header_name in dynamic_header_value["names"]:
            dynamic_body.append(
                {
                    "Column Name": simple_text(dynamic_header_name),
                }
            )
            if "description" in dynamic_header_value:
                dynamic_body[-1]["Description"] = simple_text(
                    dynamic_header_value["description"]
                )
            if "alias" in dynamic_header_value:
                dynamic_body[-1]["Alias"] = simple_text(dynamic_header_value["alias"])
    set_readme_body(
        readme_tables,
        "ReadMe for Advanced Tab",
        (POST_ADVANCED_HEADERS + PRE_ADVANCED_HEADERS)[0],
        dynamic_body,
        advanced_diffs,
    )
    return readme_tables


def json_diff(schema, old, new):
    exceptions = []
    exceptions += json_validate(schema, old)
    exceptions += json_validate(schema, new)
    basic_diffs = []
    advanced_diffs = []
    dynamic_basic_headers = OrderedDict()
    dynamic_advanced_headers = OrderedDict()
    if not exceptions:
        diffs = [
            diff_item for diff_item in traverse(schema, DiffPath(), old, new, False)
        ]
        basic_diffs = [
            DiffItemDirector.get_diff_item(
                diff_item, dynamic_basic_headers, DetailLevel.BASIC
            )
            for diff_item in filter(
                lambda diff_item: not (diff_item["is_excluded"]),
                diffs,
            )
        ]
        advanced_diffs = [
            DiffItemDirector.get_diff_item(
                diff_item, dynamic_advanced_headers, DetailLevel.ADVANCED
            )
            for diff_item in diffs
        ]
    current_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    previous_version = get_doc_id(schema, old)
    updated_version = get_doc_id(schema, new)
    tables = (
        [
            {
                "head": INFO_HEADERS,
                "title": "Info",
                "body": [
                    {
                        "Attribute": simple_text("Generated"),
                        "Value": simple_text(current_time),
                    },
                    {
                        "Attribute": simple_text("Updated Version"),
                        "Value": simple_text(updated_version),
                    },
                    {
                        "Attribute": simple_text("Previous Version"),
                        "Value": simple_text(previous_version),
                    },
                ],
            }
        ]
        + load_readme_tables(
            dynamic_basic_headers, dynamic_advanced_headers, basic_diffs, advanced_diffs
        )
        + [
            {"head": [], "title": "Basic", "body": []},
            {"head": [], "title": "Advanced", "body": []},
            {
                "head": VALIDATION_HEADERS,
                "title": "Validation",
                "body": exceptions
                if exceptions
                else [{"Issue": simple_text("No Validation issues")}],
            },
        ]
    )
    basic = next(filter(lambda item: item["title"] == "Basic", tables))
    advanced = next(filter(lambda item: item["title"] == "Advanced", tables))
    basic["head"] += PRE_BASIC_HEADERS
    advanced["head"] += PRE_ADVANCED_HEADERS
    basic["head"] += dynamic_basic_headers.keys()
    advanced["head"] += [
        header
        for header_group in dynamic_advanced_headers.values()
        for header in header_group["names"]
    ]
    for row in basic_diffs:
        row["Previous Version"] = simple_text(previous_version)
        row["Updated Version"] = simple_text(updated_version)
    basic["body"] = basic_diffs
    advanced["body"] = advanced_diffs
    basic["head"] += POST_BASIC_HEADERS
    advanced["head"] += POST_ADVANCED_HEADERS
    return tables
